# CompCertR
This is the refactored version of CompCert in the style of open simulations.

Main differences include:
(i) Proving "guarantee" conditions of each translations.
In the context of open simulations, CompCert only "rely" on it (by external call axioms) but does not "guarantee" it.
To address this gap, we needed to prove "guarantee" conditions for each per-pass proofs.
We needed to prove additional properties for injection passes only.
(ii) Changing "Callstate" of each language to carry function pointer, instead of function definition.
External module's function definition may not be visible in the local module, but it should be able to call it with function pointer.
(iii) Relax each translation's simulation relation so that first stack (main, in CompCert) may have called with arguments.
This change affects only three passes, and except for the Stackingproof the change is very small.
(iiv) Changing "step" relation to respect "symbol environment".

# CompCertM pack
"/linking" repository contains the files.
For directories that existed in CompCert (e.g. backend/), there are files that extends original CompCert's files.
New directories are roughly as follows.
bound: adequacy w.r.t. C and assembly (UpperBound.v, LowerBound.v)
compose: interaction semantics and its properties
selfsim: self-simulation of Clight and Assembly
scripts: scripts used for SLOC counting (Table 1, 2, 3)
proof: our meta theory

We list up few important definitions/proofs from the paper.

Interaction Semantics: compose/Sem.v
Module: compose/Mod.v
Module Semantics: compose/ModSem.v
RUSC: proof/RUSC.v
Compiler cocrrectness theorem (main result): driver/CompilerC.v
Parameter-MemRel: proof/SimMem.v
Parameter-SymbRel: proof/SimSymb.v
Parameter-MemPred: proof/Sound.v
parameterized open simulation: proof/SimModSem.v
open preservation: proof/Preservation.v (local_preservation_standard is the one)
AdequacyLocal: Adequacy of parameterized open simulation.
mutual-sum example: demo/mutrec
Unreadglob example: demo/unreadglob (../demo directory also contains actual optimization and proof files)
utod example: demo/utod
mixed simulation: proof/Simulation.v

For some cases, Coq formalization is slightly different from paper's presentation, but they are mathematically equivalent.

## Installation Instructions
(i) Install Paco coq library via opam (https://github.com/snu-sf/paco).
   "opam repo add coq-released https://coq.inria.fr/opam/released; opam install coq-paco"
(ii) Follow [standard CompCert build process](http://compcert.inria.fr/man/manual002.html).
   "./configure" with proper architecture.
   Install required dependencies.
   "make -j[N]"
(iii) "cd linking; make -j[N]"

Our development is tested with Coq 8.8.2.


# Unsound Assembly Examples
"unsound-asms/" includes assembly examples mentioned in Section 3.
When run on GCC-O2 and O0, it actually gives different result.

# Html Files
In "html/" directory, we put coq2html generated files for code browsing.